/*
Module : PHue++.h
Purpose: Defines the interface for a set of C++ class which encapsulate the Philips Hue Lights API as defined at
         https://developers.meethue.com/.
History: PJN / 13-01-2019 1. Initial Public release.
         PJN / 28-04-2019 1. Updated the code to clean compile on VC 2019
         PJN / 12-10-2019 1. Fixed a number of compiler warnings when the code is compiled with VS 2019 Preview
         PJN / 18-12-2019 1. Updated initialization of various structs to use C++ 11 list initialization
                          2. Fixed various Clang-Tidy static code analysis warnings in the code.
         PJN / 01-03-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 07-11-2021 1. Updated copyright details.
                          2. Fixed some compiler warnings from VC 2019 & VC 2022
         PJN / 05-06-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations.

Copyright (c) 2019 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#ifndef __PHUEPP_H__
#define __PHUEPP_H__


//////////////////// Includes /////////////////////////////////////////////////

#include "WinHTTPWrappers.h"
#include "JSON++.h"

#ifndef _STRING_
#pragma message("To avoid this message, please put string in your pre compiled header (normally stdafx.h)")
#include <string>
#endif //#ifndef _STRING_

#ifndef _ARRAY_
#pragma message("To avoid this message, please put array in your pre compiled header (normally stdafx.h)")
#include <array>
#endif //#ifndef _ARRAY_

#ifndef _OPTIONAL_
#pragma message("To avoid this message, please put optional in your pre compiled header (normally stdafx.h)")
#include <optional>
#endif //#ifndef _OPTIONAL_

#ifndef _CMATH_
#pragma message("To avoid this message, please put cmath in your pre compiled header (normally stdafx.h)")
#include <cmath>
#endif //#ifndef _CMATH_

#ifndef __ATLSTR_H__
#pragma message("To avoid this message, please put atlstr.h in your pre compiled header (normally stdafx.h)")
#include <atlfile.h>
#endif //#ifndef __ATLSTR_H__


//////////////////// Classes //////////////////////////////////////////////////

namespace PhilipsHue
{

//Enums
enum class ERROR_TYPE
{
  ZERO_ERROR = 0,
  UNAUTHORIZED_USER = 1,
  BODY_CONTAINS_INVALID_JSON = 2,
  RESOURCE_NOT_AVAILABLE = 3,
  METHOD_NOT_AVAILABLE = 4,
  MISSING_PARAMETERS = 5,
  PARAMETER_NOT_AVAILABLE = 6,
  INVALID_VALUE = 7,
  PARAMETER_IS_NOT_MODIFIABLE_1 = 8,
  TOO_MANY_ITEMS_IN_LIST = 11,
  PORTAL_CONNECTION_REQUIRED = 12,
  INTERNAL_ERROR = 901,
  LINK_BUTTON_NOT_PRESSED = 101,
  DHCP_CANNOT_BE_DISABLED = 110,
  INVALID_UPDATESTATE = 111,
  PARAMETER_IS_NOT_MODIFIABLE_2 = 201,
  COMMISSIONABLE_LIGHT_LIST_FULL = 203,
  GROUP_COULD_NOT_BE_CREATED = 301,
  DEVICE_COULD_NOT_BE_ADDED_TO_GROUP = 302,
  DEVICE_COULD_NOT_BE_ADDED_TO_SCENE = 304,
  UPDATE_OR_DELETE_IS_NOT_ALLOWED = 305,
  LIGHT_IS_ALREADY_IN_ANOTHER_ROOM = 306,
  SCENE_COULD_NOT_BE_CREATED_1 = 401,
  SCENE_COULD_NOT_BE_CREATED_2 = 402,
  SCENE_COULD_NOT_BE_REMOVED = 403,
  NOT_ALLOWED_TO_CREATE_SENSOR = 501,
  SENSOR_LIST_IS_FULL = 502,
  COMMISSIONABLE_SENSOR_LIST_FULL = 503,
  RULE_ENGINE_FULL = 601,
  CONDITION_ERROR = 607,
  ACTION_ERROR = 608,
  UNABLE_TO_ACTIVATE = 609,
  SCHEDULE_LIST_IS_FULL = 701,
  SCHEDULE_TIMEZONE_NOT_VALID = 702,
  SCHEDULE_CANNOT_SET_TIME = 703,
  CANNOT_CREATE_SCHEDULE = 704,
  CANNOT_ENABLE_SCHEDULE = 705,
  COMMAND_ERROR = 706,
  SOURCE_MODEL_INVALID = 801,
  SOURCE_FACTORY_NEW = 802,
  INVALID_STATE = 803
};


//An individual error response
class CErrorResponse
{
public:
//Member variables
  ERROR_TYPE m_Type{ERROR_TYPE::ZERO_ERROR};
  std::wstring m_sAddress;
  std::wstring m_sDescription;
};


//Helper method to parse UPnP headers (used by the UPnPDiscover function)
std::string ParseHTTPHeader(const std::string& sHeader, LPCSTR pszKey)
{
  std::string sValue;
  auto nFindStart{sHeader.find(pszKey)};
  if (nFindStart != std::string::npos)
  {
    nFindStart += strlen(pszKey);
    const auto nFindEnd{sHeader.find("\r\n", nFindStart)};
    if (nFindEnd != std::string::npos)
      sValue = sHeader.substr(nFindStart, nFindEnd - nFindStart);
    else
      sValue = sHeader.substr(nFindStart);
  }
  return sValue;
}


//Converts a color temperature in kelvin to reciprocal megakelvin aka mirek
constexpr int32_t KelvinToMired(int32_t kelvin) noexcept
{
  return static_cast<int32_t>(0.5f + (1000000.0 / kelvin));
}


//Converts a color temperature in reciprocal megakelvin aka mirek to Kelvin
constexpr int32_t MiredToKelvin(int32_t mired) noexcept
{
  return static_cast<int32_t>(0.5f + (1000000.0 / mired));
}


//The GamutA triangle cartesian points
static std::array<std::pair<float, float>, 3> g_GamutA
{ {
  { 0.704f, 0.296f }, { 0.2151f, 0.7106f }, { 0.138f, 0.08f }
} };


//The GamutB triangle cartesian points
static std::array<std::pair<float, float>, 3> g_GamutB
{ {
  { 0.675f, 0.322f }, { 0.4091f, 0.518f }, { 0.167f, 0.04f }
} };


//The GamutC triangle cartesian points
std::array<std::pair<float, float>, 3> g_GamutC
{ {
  { 0.692f, 0.308f }, { 0.17f, 0.7f }, { 0.153f, 0.048f }
} };


//Calculate the cross product of two cartesian points
constexpr float CrossProduct(const std::pair<float, float>& p1, const std::pair<float, float>& p2) noexcept
{
  return ((p1.first * p2.second) - (p1.second * p2.first));
}


//Is the cartesian coordinate "point" inside the "gamut" triangle
bool PointInLampsReach(const std::array<std::pair<float, float>, 3>& gamut, const std::pair<float, float>& point) noexcept
{
#pragma warning(suppress: 26446)
  const std::pair<float, float>& red{gamut[0]};
#pragma warning(suppress: 26446)
  const std::pair<float, float>& green{gamut[1]};
#pragma warning(suppress: 26446)
  const std::pair<float, float>& blue{gamut[2]};
  const std::pair<float, float> v1{green.first - red.first, green.second - red.second};
  const std::pair<float, float> v2{blue.first - red.first, blue.second - red.second};
  const std::pair<float, float> q{point.first - red.first, point.second - red.second};
  const float s{CrossProduct(q, v2) / CrossProduct(v1, v2)};
  const float t{CrossProduct(v1, q) / CrossProduct(v1, v2)};

  return (s >= 0.0f) && (t >= 0.0f) and ((s + t) <= 1.0f);
}


//Find the distance between two cartesian points
float DistanceBetweenTwoPoints(const std::pair<float, float>& p1, const std::pair<float, float>& p2) noexcept
{
  const float dx{p1.first - p2.first};
  const float dy{p1.second - p2.second};
  return sqrt((dx * dx) + (dy *dy));
}


//Find the closest point on the line through cartesian points "A" and "B" to the cartesian point "P"
std::pair<float, float> ClosestPointToLine(const std::pair<float, float>& A, const std::pair<float, float>& B, const std::pair<float, float>& P) noexcept
{
  const std::pair<float, float> AP{P.first - A.first, P.second - A.second};
  const std::pair<float, float> AB{B.first - A.first, B.second - A.second};
  const float ab2{(AB.first * AB.first) + (AB.second * AB.second)};
  const float ap_ab{(AP.first * AB.first) + (AP.second * AB.second)};
  float t = ap_ab / ab2;

  if (t < 0.0f)
    t = 0.0f;
  else if (t > 1.0f)
    t = 1.0f;

  return std::pair<float, float>{A.first + (AB.first * t), A.second + (AB.second * t)};
}


//Find the closest point on the "gamut" triangle to the cartesian point "point"
std::pair<float, float> ClosestPointOnTriangle(const std::array<std::pair<float, float>, 3>& gamut, const std::pair<float, float>& point) noexcept
{
#pragma warning(suppress: 26446)
  const std::pair<float, float>& red{gamut[0]};
#pragma warning(suppress: 26446)
  const std::pair<float, float>& green{gamut[1]};
#pragma warning(suppress: 26446)
  const std::pair<float, float>& blue{gamut[2]};
  const std::pair<float, float> pAB{ClosestPointToLine(red, green, point)};
  const std::pair<float, float> pAC{ClosestPointToLine(blue, red, point)};
  const std::pair<float, float> pBC{ClosestPointToLine(green, blue, point)};
  const float dAB{DistanceBetweenTwoPoints(point, pAB)};
  const float dAC{DistanceBetweenTwoPoints(point, pAC)};
  const float dBC{DistanceBetweenTwoPoints(point, pBC)};

  float lowest{dAB};
  std::pair<float, float> closest_point{pAB};
  if (dAC < lowest)
  {
    lowest = dAC;
    closest_point = pAC;
  }
  if (dBC < lowest)
    closest_point = pBC;
  return closest_point;
}


//Converts from RGB color space to xy color space. This code is based on the Python code at https://github.com/benknight/hue-python-rgb-converter
std::pair<float, float> RGBToXY(COLORREF color, const std::array<std::pair<float, float>, 3>& gamut = g_GamutB) noexcept
{
  //Step 1, get the normalized RGB values
  float red{GetRValue(color) / 255.0f};
  float green{GetGValue(color) / 255.0f};
  float blue{GetBValue(color) / 255.0f};

  //Step 2, Apply a gamma correction
  red = (red > 0.04045f) ? pow((red + 0.055f) / (1.0f + 0.055f), 2.4f) : (red / 12.92f);
  green = (green > 0.04045f) ? pow((green + 0.055f) / (1.0f + 0.055f), 2.4f) : (green / 12.92f);
  blue = (blue > 0.04045f) ? pow((blue + 0.055f) / (1.0f + 0.055f), 2.4f) : (blue / 12.92f);

  //Step 3, Convert to XYZ using the Wide RGB D65 conversion formula
  const float X{(red * 0.664511f) + (green * 0.154324f) + (blue * 0.162028f)};
  const float Y{(red * 0.283881f) + (green * 0.668433f) + (blue * 0.047685f)};
  const float Z{(red * 0.000088f) + (green * 0.072310f) + (blue * 0.986039f)};

  //Step 4, Calculate the xy values from the XYZ values
  std::pair<float, float> xy{X / (X + Y + Z), Y / (X + Y + Z)};

  //Step 5, Check if the found "xy" value is within the color gamut
  if (!PointInLampsReach(gamut, xy))
  {
    //Step 6, calculate the closest point on the color "gamut" triangle to "xy"
    return ClosestPointOnTriangle(gamut, xy);
  }
  else
  {
    //Step 7, return xy
    return xy;
  }
}


//Converts from xy color space to RGB color space. This code is based on the Python code at https://github.com/benknight/hue-python-rgb-converter
COLORREF XYToRGB(float x, float y, float brightness = 1, const std::array<std::pair<float, float>, 3>& gamut = g_GamutB) noexcept
{
  //Step 1. Check the xy value is within the color gamut of the lamp
  std::pair<float, float> xy{x, y};
  if (!PointInLampsReach(gamut, xy))
  {
    //Step 2, calculate the most accurate color the given light can actually do
    xy = ClosestPointOnTriangle(gamut, xy);
  }

  //Step 3, calculate XYZ values
  const float z{1.0f - xy.first - xy.second};
  const float Y{brightness}; //The given brightness value
  const float X{(Y / xy.second) * xy.first};
  const float Z{(Y / xy.second) * z};

  //Step 2, Convert to RGB using the sRGB D65 conversion
  float r{(X * 1.656492f) - (Y * 0.354851f) - (Z * 0.255038f)};
  float g{(-X * 0.707196f) + (Y * 1.655397f) + (Z * 0.036152f)};
  float b{(X * 0.051713f) - (Y * 0.121364f) + (Z * 1.011530f)};

  //Step 3, Apply reverse gamma correction
  r = r <= 0.0031308f ? 12.92f * r : (1.0f + 0.055f) * pow(r, (1.0f / 2.4f)) - 0.055f;
  g = g <= 0.0031308f ? 12.92f * g : (1.0f + 0.055f) * pow(g, (1.0f / 2.4f)) - 0.055f;
  b = b <= 0.0031308f ? 12.92f * b : (1.0f + 0.055f) * pow(b, (1.0f / 2.4f)) - 0.055f;

  //Step 4, Bring all negative components to zero
  if (r < 0.0f)
    r = 0.0f;
  if (g < 0.0f)
    g = 0.0f;
  if (b < 0.0f)
    b = 0.0f;

  //Step 5, If one component is greater than 1, weight components by that value
  float max_value{r};
  if (g > max_value)
    max_value = g;
  if (b > max_value)
    max_value = b;
  if (max_value > 1.0f)
  {
    r = r / max_value;
    g = g / max_value;
    b = b / max_value;
  }

#pragma warning(suppress: 26467)
  return RGB(static_cast<BYTE>(0.5f + (r * 255)), static_cast<BYTE>(0.5f + (g * 255)), static_cast<BYTE>(0.5f + (b * 255)));
}


//Converts from the RGB color space to the HSL color space. This code is based on the JavaScript code at https://stackoverflow.com/questions/2353211/hsl-to-rgb-color-conversion
constexpr std::array<float, 3> RGBToHSL(COLORREF color) noexcept
{
  //Step 1, get the normalized RGB values
  float r{GetRValue(color) / 255.0f};
  float g{GetGValue(color) / 255.0f};
  float b{GetBValue(color) / 255.0f};

  //Step 2. Find the max and min values
  float max{r};
  if (g > max)
    max = g;
  if (b > max)
    max = b;
  float min{r};
  if (g < min)
    min = g;
  if (b < min)
    min = b;

  //Step 3. Calculate the actual HSL values
  std::array<float, 3> hsl{ { 0.0f, 0.0f, 0.0f } };
#pragma warning(suppress: 26446)
  hsl[2] = (max + min) / 2;
  if (max == min)
  {
#pragma warning(suppress: 26446)
    hsl[0] = 0;
#pragma warning(suppress: 26446)
    hsl[1] = 0;
  }
  else
  {
    float d{max - min};
#pragma warning(suppress: 26446)
    hsl[1] = hsl[2] > 0.5 ? d / (2 - max - min) : d / (max + min);
    if (max == r)
#pragma warning(suppress: 26446)
      hsl[0] = (g - b) / d + (g < b ? 6.0f : 0.0f);
    else if (max == g)
#pragma warning(suppress: 26446)
      hsl[0] = (b - r) / d + 2.0f;
    else
#pragma warning(suppress: 26446)
      hsl[0] = (r - g) / d + 4.0f;

#pragma warning(suppress: 26446)
    hsl[0] /= 6.0f;
  }

  return hsl;
}


//Helper function used by RGBToHSL function
constexpr float HueToRgb(float p, float q, float t) noexcept
{
  if (t < 0.0f)
    t += 1;
  if (t > 1.0f)
    t -= 1;
  if (t < (1.0f / 6.0f))
    return p + (q - p) * 6.0f * t;
  if (t < (1.0f / 2.0f))
    return q;
  if (t < (2.0f / 3.0f))
    return p + (q - p) * ((2.0f / 3.0f) - t) * 6.0f;
  return p;
}


//Converts from the HSL color space to the RGB color space. This code is based on the JavaScript code at https://stackoverflow.com/questions/2353211/hsl-to-rgb-color-conversion
constexpr COLORREF HSLToRGB(float h, float s, float l) noexcept
{
  float r{0};
  float g{0};
  float b{0};

  if (s == 0)
  {
    //Achromatic
    r = l;
    g = l;
    b = l;
  }
  else
  {
    const float q{l < 0.5f ? l * (1 + s) : l + s - l * s};
    const float p{2 * l - q};
    r = HueToRgb(p, q, h + (1.0f / 3.0f));
    g = HueToRgb(p, q, h);
    b = HueToRgb(p, q, h - (1.0f / 3.0f));
  }

#pragma warning(suppress: 26467)
  return RGB(min(floor(r*256), 255), min(floor(g * 256), 255), min(floor(b * 256), 255));
}


//UPnP discovery function
HRESULT UPnPDiscover(_Inout_ std::vector<std::pair<std::string, std::string>>& devices, DWORD dwTimeout = 5, LPCSTR pszSearchTarget = "ssdp:all", LPCSTR pszUserAgent = nullptr, HANDLE hCancel = nullptr)
{
  //Reset the output parameter
  devices.clear();

  //Create the UDP IPv4 datagram socket
  const SOCKET hSocket{socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)};
  if (hSocket == INVALID_SOCKET)
    return ATL::AtlHresultFromWin32(WSAGetLastError());

  //Set the socket to non-blocking mode
  DWORD dwNonBlocking{1};
  if (ioctlsocket(hSocket, FIONBIO, &dwNonBlocking) == SOCKET_ERROR)
  {
    const int nError{WSAGetLastError()};
    closesocket(hSocket);
    return ATL::AtlHresultFromWin32(nError);
  }

  //Set the socket to broadcast mode
  constexpr const BOOL bBroadcast{TRUE};
#pragma warning(suppress: 26490)
  if  (setsockopt(hSocket, SOL_SOCKET, SO_BROADCAST, reinterpret_cast<const char*>(&bBroadcast), sizeof(bBroadcast)) == SOCKET_ERROR)
  {
    const int nError{WSAGetLastError()};
    closesocket(hSocket);
    return ATL::AtlHresultFromWin32(nError);
  }

  //Send the UPnP discover request to the IPv4 limited broadcast address
  SOCKADDR_IN destAddr{};
  destAddr.sin_family = AF_INET;
  destAddr.sin_port = htons(1900);
  destAddr.sin_addr.S_un.S_addr = INADDR_BROADCAST;
  std::ostringstream ss;
  ss << "M-SEARCH * HTTP/1.1\r\nHOST: 239.255.255.250:1900\r\nMAN: \"ssdp:discover\"\r\nMX: ";
  ss << dwTimeout;
  ss << "\r\nST: ";
  ss << pszSearchTarget;
  if (pszUserAgent != nullptr)
  {
    ss << "USER-AGENT: ";
    ss << pszUserAgent;
    ss << "\r\n";
  }
  ss << "\r\n";
  std::string sDiscover{ss.str()};
#pragma warning(suppress: 26472 26490)
  if (sendto(hSocket, sDiscover.c_str(), static_cast<int>(sDiscover.length()), 0, reinterpret_cast<const SOCKADDR*>(&destAddr), sizeof(destAddr)) == SOCKET_ERROR)
  {
    const int nError{WSAGetLastError()};
    closesocket(hSocket);
    return ATL::AtlHresultFromWin32(nError);
  }

  //Loop around reading the responses
  bool bContinue{true};
#pragma warning(suppress: 26489)
  std::string sResponse;
  const ULONGLONG dwStartTick{GetTickCount64()};
  while (bContinue)
  {
    //Read something from the socket
    std::array<char, 4096> receiveBuffer{};
    SOCKADDR_STORAGE sockAddr2{};
    int nSockAddrLen{sizeof(sockAddr2)};
#pragma warning(suppress: 26472 26490)
    const int nReceived{recvfrom(hSocket, receiveBuffer.data(), static_cast<int>(receiveBuffer.size()), 0, reinterpret_cast<SOCKADDR*>(&sockAddr2), &nSockAddrLen)};
    if (nReceived > 0)
#pragma warning(suppress: 26481 26486)
      sResponse.insert(sResponse.end(), receiveBuffer.data(), receiveBuffer.data() + nReceived);
    else if (nReceived == SOCKET_ERROR && (WSAGetLastError() != WSAEWOULDBLOCK))
    {
      const int nError{WSAGetLastError()};
      closesocket(hSocket);
      return ATL::AtlHresultFromWin32(nError);
    }

    //terminate the loop or prepare for the next time around
    if ((GetTickCount64() - dwStartTick) > (static_cast<ULONGLONG>(dwTimeout) * 1000))
      bContinue = false;
    else if ((hCancel != nullptr) && (WaitForSingleObject(hCancel, 0) == WAIT_OBJECT_0))
      bContinue = false;
    else
      Sleep(0); //Yield our timeslice before we loop around
  }

  //Close the socket now that we are finished with it
  closesocket(hSocket);

  //Parse the response blob into an array of string responses
  std::vector<std::string> stringResponses;
  size_t nPrevFind{0};
  auto nCurrentFind{sResponse.find("\r\n\r\n")};
  while (nCurrentFind != std::string::npos)
  {
#pragma warning(suppress: 26489)
    stringResponses.push_back(sResponse.substr(nPrevFind, nCurrentFind - nPrevFind));

    //Prepare for the next time around
    nCurrentFind += 4;
    nPrevFind = nCurrentFind;
#pragma warning(suppress: 26489)
    nCurrentFind = sResponse.find("\r\n\r\n", nCurrentFind);
  }

  //Now convert the array of string responses into an array of server / location pairs
  for (const auto& response : stringResponses)
  {
    std::pair<std::string, std::string> device;
#pragma warning(suppress: 26486 26489)
    device.first = ParseHTTPHeader(response, "LOCATION: ");
    if (device.first.length() == 0)
#pragma warning(suppress: 26486 26489)
      device.first = ParseHTTPHeader(response, "LOCATION:");
#pragma warning(suppress: 26486 26489)
    device.second = ParseHTTPHeader(response, "SERVER: ");
    if (device.second.length() == 0)
#pragma warning(suppress: 26486 26489)
      device.second = ParseHTTPHeader(response, "SERVER:");
#pragma warning(suppress: 26486 26489)
    if (ParseHTTPHeader(response, "NTS: ") != "ssdp::byebye") //Ignore byebye notifications
    {
#pragma warning(suppress: 26489)
      if (device.first.length() && std::find_if(devices.begin(), devices.end(), 
                                                [&](const std::pair<std::string, std::string>& item) noexcept
                                                {
                                                  return item.first == device.first;
#pragma warning(suppress: 26489)
                                                }) == devices.end())
      {
#pragma warning(suppress: 26489)
        devices.push_back(device);
      }
    }
  }

  return S_OK;
}


class CLightState
{
public:

//Enums
  enum class ALERT
  {
    NOT_SET,
    NONE,
    SELECT,
    LSELECT
  };

  enum class EFFECT
  {
    NOT_SET,
    NONE,
    COLORLOOP
  };

//Member variables
  std::optional<bool> m_On;
  std::optional<uint8_t> m_Brightness;
  std::optional<uint16_t> m_Hue;
  std::optional<uint8_t> m_Saturation;
  std::optional<std::pair<float, float>> m_XY;
  std::optional<uint16_t> m_CT;
  ALERT m_Alert{ALERT::NOT_SET};
  EFFECT m_Effect{EFFECT::NOT_SET};
  std::optional<uint16_t> m_TransitionTime;
  std::optional<int16_t> m_BrightnessInc;
  std::optional<int16_t> m_SaturationInc;
  std::optional<int32_t> m_HueInc;
  std::optional<int32_t> m_CTInc;
  std::optional<std::pair<float, float>> m_XYInc;
};


//Forward reference
class CBridge;


//Class which represents a Philips Hue light
class CLight
{
public:
//Methods
  HRESULT Set(_In_ const CLightState& state, _Inout_ std::vector<CErrorResponse>& errors);

  HRESULT SetOn(_Inout_ std::vector<CErrorResponse>& errors)
  {
    CLightState state;
    state.m_On = true;
#pragma warning(suppress: 26486)
    return Set(state, errors);
  }

  HRESULT SetOff(_Inout_ std::vector<CErrorResponse>& errors)
  {
    CLightState state;
    state.m_On = false;
#pragma warning(suppress: 26486)
    return Set(state, errors);
  }

  HRESULT SetBrightness(_In_ uint8_t brightness, _Inout_ std::vector<CErrorResponse>& errors)
  {
    CLightState state;
    state.m_Brightness = brightness;
#pragma warning(suppress: 26486)
    return Set(state, errors);
  }

  HRESULT SetHue(_In_ uint16_t hue, _Inout_ std::vector<CErrorResponse>& errors)
  {
    CLightState state;
    state.m_Hue = hue;
#pragma warning(suppress: 26486)
    return Set(state, errors);
  }

  HRESULT SetSaturation(_In_ uint8_t saturation, _Inout_ std::vector<CErrorResponse>& errors)
  {
    CLightState state;
    state.m_Saturation = saturation;
#pragma warning(suppress: 26486)
    return Set(state, errors);
  }

  HRESULT SetColorTemperature(_In_ uint16_t colorTemp, _Inout_ std::vector<CErrorResponse>& errors)
  {
    CLightState state;
    state.m_CT = colorTemp;
#pragma warning(suppress: 26486)
    return Set(state, errors);
  }

  HRESULT SetAlert(_In_ CLightState::ALERT alert, _Inout_ std::vector<CErrorResponse>& errors)
  {
    CLightState state;
    state.m_Alert = alert;
#pragma warning(suppress: 26486)
    return Set(state, errors);
  }

  HRESULT SetEffect(_In_ CLightState::EFFECT effect, _Inout_ std::vector<CErrorResponse>& errors)
  {
    CLightState state;
    state.m_Effect = effect;
#pragma warning(suppress: 26486)
    return Set(state, errors);
  }

  HRESULT SetXY(_In_ const std::pair<float, float>& xy, _Inout_ std::vector<CErrorResponse>& errors)
  {
    CLightState state;
    state.m_XY = xy;
#pragma warning(suppress: 26486)
    return Set(state, errors);
  }

  HRESULT Delete(_Inout_ std::vector<CErrorResponse>& errors);
  HRESULT Rename(_In_z_ LPCSTR pszNewName, _Inout_ std::vector<CErrorResponse>& errors);
  HRESULT Refresh(_Inout_ std::vector<CErrorResponse>& errors);

  [[nodiscard]] int ID() const noexcept
  {
    return m_nID;
  }

  HRESULT GetOn(_Out_ bool& bOn)
  {
    return BooleanParam(m_LastState, L"/state/on", bOn);
  }

  HRESULT GetBrightness(_Out_ uint8_t& brightness)
  {
    return NumberParam<uint8_t>(m_LastState, L"/state/bri", brightness);
  }

  HRESULT GetHue(_Out_ uint16_t& hue)
  {
    return NumberParam<uint16_t>(m_LastState, L"/state/hue", hue);
  }

  HRESULT GetSaturation(_Out_ uint8_t& saturation)
  {
    return NumberParam<uint8_t>(m_LastState, L"/state/sat", saturation);
  }

  HRESULT GetColorTemperature(_Out_ uint16_t& colorTemp)
  {
    return NumberParam<uint16_t>(m_LastState, L"/state/ct", colorTemp);
  }

  HRESULT GetReachable(_Out_ bool& bReachable)
  {
    return BooleanParam(m_LastState, L"/state/reachable", bReachable);
  }

  HRESULT GetColorMode(_Out_ std::wstring& sColorMode)
  {
    return StringParam(m_LastState, L"/state/colormode", sColorMode);
  }

  HRESULT GetAlert(_Out_ CLightState::ALERT& alert)
  {
    std::wstring sAlert;
    const HRESULT hr{StringParam(m_LastState, L"/state/alert", sAlert)};
    if (FAILED(hr))
      return hr;
    if (sAlert == L"none")
      alert = CLightState::ALERT::NONE;
    else if (sAlert == L"select")
      alert = CLightState::ALERT::SELECT;
    else if (sAlert == L"lselect")
      alert = CLightState::ALERT::LSELECT;
    else
      alert = CLightState::ALERT::NOT_SET;
    return S_OK;
  }

  HRESULT GetEffect(_Out_ CLightState::EFFECT& effect)
  {
    std::wstring sEffect;
    const HRESULT hr{StringParam(m_LastState, L"/state/effect", sEffect)};
    if (FAILED(hr))
      return hr;
    if (sEffect == L"none")
      effect = CLightState::EFFECT::NONE;
    else if (sEffect == L"colorloop")
      effect = CLightState::EFFECT::COLORLOOP;
    else
      effect = CLightState::EFFECT::NOT_SET;
    return S_OK;
  }

  HRESULT GetXY(_Out_ std::pair<float, float>& xy)
  {
    try
    {
      JSONPP::CValue* pChild{m_LastState.JSONPointer(L"/state/xy")};
      if (pChild == nullptr)
        return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
      const JSONPP::Array& xyVal{pChild->AsArray()};
      if (xyVal.size() < 2)
        return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
#pragma warning(suppress: 26446)
      xy.first = static_cast<float>(xyVal[0].AsNumber());
#pragma warning(suppress: 26446)
      xy.second = static_cast<float>(xyVal[1].AsNumber());
    }
    catch (JSONPP::CParseException&)
    {
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
    }
    return S_OK;
  }

  HRESULT GetType(_Out_ std::wstring& sType)
  {
    return StringParam(m_LastState, L"/type", sType);
  }

  HRESULT GetModelID(_Out_ std::wstring& sModelID)
  {
    return StringParam(m_LastState, L"/modelid", sModelID);
  }

  HRESULT GetName(_Out_ std::wstring& sName)
  {
    return StringParam(m_LastState, L"/name", sName);
  }

  HRESULT GetUniqueID(_Out_ std::wstring& sUniqueID)
  {
    return StringParam(m_LastState, L"/uniqueid", sUniqueID);
  }

  HRESULT GetManufacturerName(_Out_ std::wstring& sManufacturerName)
  {
    return StringParam(m_LastState, L"/manufacturername", sManufacturerName);
  }

  HRESULT GetLuminaireUniqueID(_Out_ std::wstring& sLuminaireUniqueID)
  {
    return StringParam(m_LastState, L"/luminaireuniqueid", sLuminaireUniqueID);
  }

//Member variables
  JSONPP::CValue m_LastResponse; //The last response of the light
  JSONPP::CValue m_LastState; //The most recent state of the light

protected:
//Methods
  static HRESULT BooleanParam(_In_ JSONPP::CValue& response, _In_z_ LPCWSTR pszJSONPointer, _Out_ bool& bParam)
  {
    try
    {
      JSONPP::CValue* pChild{response.JSONPointer(pszJSONPointer)};
      if (pChild == nullptr)
        return ATL::AtlHresultFromWin32(ERROR_NOT_FOUND);
      bParam = pChild->AsBoolean();
    }
    catch (JSONPP::CParseException&)
    {
      return ATL::AtlHresultFromWin32(ERROR_NOT_FOUND);
    }
    return S_OK;
  }

  static HRESULT StringParam(_In_ JSONPP::CValue& response, _In_z_ LPCWSTR pszJSONPointer, _Out_ std::wstring& sParam)
  {
    try
    {
      JSONPP::CValue* pChild{response.JSONPointer(pszJSONPointer)};
      if (pChild == nullptr)
        return ATL::AtlHresultFromWin32(ERROR_NOT_FOUND);
      sParam = pChild->AsString();
    }
    catch (JSONPP::CParseException&)
    {
      return ATL::AtlHresultFromWin32(ERROR_NOT_FOUND);
    }
    return S_OK;
  }

  template<typename ParamType>
  static HRESULT NumberParam(_In_ JSONPP::CValue& response, _In_z_ LPCWSTR pszJSONPointer, _Out_ ParamType& bParam)
  {
    try
    {
      JSONPP::CValue* pChild{response.JSONPointer(pszJSONPointer)};
      if (pChild == nullptr)
        return ATL::AtlHresultFromWin32(ERROR_NOT_FOUND);
      bParam = static_cast<ParamType>(pChild->AsNumber());
    }
    catch (JSONPP::CParseException&)
    {
      return ATL::AtlHresultFromWin32(ERROR_NOT_FOUND);
    }
    return S_OK;
  }


//Member variables
  int m_nID{0}; //The ID of the light
  CBridge* m_pBridge{nullptr}; //The owning bridge reference

  friend class CBridge;
};


//Main class which represents the Philips Hue bridge
class CBridge
{
public:
//Methods
  HRESULT Initialize(_In_ const WinHTTPWrappers::CSession& session, _In_z_ LPCWSTR pwszServerName, _In_ INTERNET_PORT nServerPort = INTERNET_DEFAULT_PORT) noexcept
  {
    return m_Connection.Initialize(session, pwszServerName, nServerPort);
  }

  HRESULT GetDescriptionXML(_Out_ std::string& sResponse)
  {
    return Get(L"/description.xml", sResponse);
  }

  static HRESULT IsErrorResponse(_In_ const std::string& sResponse, _Inout_ bool& bError, _Inout_ std::vector<CErrorResponse>& errors)
  {
    //Default the output parameters
    errors.clear();
    bError = false;

    //Parse the JSON
    JSONPP::CValue value;
    try
    {
      value.Parse(sResponse.c_str());
    }
    catch (JSONPP::CParseException&)
    {
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
    }

    //And check to see if there are any "error" objects in it
    if (value.IsArray())
    {
      JSONPP::Array& arr{value.AsArray()};
      for (auto& sub : arr)
      {
        if (sub.IsObject())
        {
          JSONPP::Object& object{sub.AsObject()};
          const auto iter{object.find(L"error")};
          if (iter != object.cend())
          {
            bError = true;
            if (iter->second.IsObject())
            {
              JSONPP::Object& sub2{iter->second.AsObject()};
              CErrorResponse err;
              err.m_Type = static_cast<ERROR_TYPE>(static_cast<int>(sub2[L"type"].AsNumber()));
#pragma warning(suppress: 26489)
              err.m_sAddress = sub2[L"address"].AsString();
#pragma warning(suppress: 26489)
              err.m_sDescription = sub2[L"description"].AsString();
#pragma warning(suppress: 26489)
              errors.push_back(err);
            }
          }
        }
      }
    }

    return S_OK;
  }

  HRESULT GetLights(_Inout_ std::vector<CErrorResponse>& errors)
  {
    //Validate our parameters
    ASSERT(m_sClientKey.length());

    std::wostringstream ss;
    ss << L"/api/";
    ss << m_sClientKey;
    ss << L"/lights";
    std::wstring sURL(ss.str());

    std::string sResponse;
    HRESULT hr{Get(sURL.c_str(), sResponse)};
    if (FAILED(hr))
      return hr;
    try
    {
      m_LastResponse.Parse(sResponse.c_str());
    }
    catch (JSONPP::CParseException&)
    {
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
    }

    bool bErrorResponse{false};
    hr = IsErrorResponse(sResponse, bErrorResponse, errors);
    if (FAILED(hr) || bErrorResponse)
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
    if (!m_LastResponse.IsObject())
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
    m_LastLights = m_LastResponse;
    return S_OK;
  }

  bool HasLight(_In_ int nID)
  {
    std::wostringstream ss;
    ss << nID;
    return m_LastLights.HasChild(ss.str().c_str());
  }

  HRESULT GetLight(_In_ int nID, _Inout_ CLight& light)
  {
    std::wostringstream ss;
    ss << L"/";
    ss << nID;
    std::wstring sID{ss.str()};
    try
    {
      const JSONPP::CValue* pLight{m_LastLights.JSONPointer(sID.c_str())};
      if (pLight == nullptr)
        return S_FALSE;
    }
    catch (JSONPP::CParseException&)
    {
      return ATL::AtlHresultFromWin32(ERROR_NOT_FOUND);
    }

    //Update the values in the output parameter
    light.m_nID = nID;
    light.m_pBridge = this;
    return S_OK;
  }

  HRESULT GetNewLights(_Inout_ std::vector<int>& newLights, _Inout_ std::wstring& sLastScan, _Inout_ std::vector<CErrorResponse>& errors)
  {
    //Validate our parameters
    ASSERT(m_sClientKey.length());

    //Reset the output parameters
    newLights.clear();
    sLastScan.clear();

    std::wostringstream ss;
    ss << L"/api/";
    ss << m_sClientKey;
    ss << L"/lights/new";
    std::wstring sURL{ss.str()};
    std::string sResponse;
    HRESULT hr{Get(sURL.c_str(), sResponse)};
    if (FAILED(hr))
      return hr;
    try
    {
      m_LastResponse.Parse(sResponse.c_str());
    }
    catch (JSONPP::CParseException&)
    {
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
    }
    bool bErrorResponse{false};
    hr = IsErrorResponse(sResponse, bErrorResponse, errors);
    if (FAILED(hr) || bErrorResponse)
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);

    //Extract the new lights and lastscan value from the response
    if (!m_LastResponse.IsObject())
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
    JSONPP::Object& obj{m_LastResponse.AsObject()};
    for (auto& sub : obj)
    {
#pragma warning(suppress: 26489)
      if (sub.first == L"lastscan")
#pragma warning(suppress: 26489)
        sLastScan = sub.second.AsString();
      else
#pragma warning(suppress: 26489)
        newLights.push_back(_wtoi(sub.first.c_str()));
    }

    return S_OK;
  }

  HRESULT SearchNewLights(_Inout_ std::vector<CErrorResponse>& errors)
  {
    //Validate our parameters
    ASSERT(m_sClientKey.length());

    std::wostringstream ss;
    ss << L"/api/";
    ss << m_sClientKey;
    ss << L"/lights";
    std::wstring sURL{ss.str()};
    std::string sResponse;
    HRESULT hr{Post(sURL.c_str(), nullptr, sResponse)}; //Note currently we do not support a request body for Search for new lights
    if (FAILED(hr))
      return hr;
    try
    {
      m_LastResponse.Parse(sResponse.c_str());
    }
    catch (JSONPP::CParseException&)
    {
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
    }
    bool bErrorResponse{false};
    hr = IsErrorResponse(sResponse, bErrorResponse, errors);
    if (FAILED(hr) || bErrorResponse)
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
    return S_OK;
  }

  HRESULT CreateNewUser(_In_z_ LPCSTR pszUserName, _Inout_ std::wstring& sClientKey, _Inout_ std::vector<CErrorResponse>& errors)
  {
    std::ostringstream ss;
    ss << R"({"devicetype": ")";
    ss << pszUserName;
    ss << "\"}";
    std::string sRequest{ss.str()};
    std::string sResponse;
    HRESULT hr{Post(L"/api", ss.str().c_str(), sResponse)};
    if (FAILED(hr))
      return hr;
    try
    {
      m_LastResponse.Parse(sResponse.c_str());
    }
    catch (JSONPP::CParseException&)
    {
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
    }
    bool bErrorResponse{false};
    hr = IsErrorResponse(sResponse, bErrorResponse, errors);
    if (FAILED(hr) || bErrorResponse)
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);

    //Extract the client key from the response
    if (!m_LastResponse.IsArray())
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
    JSONPP::Array& arr{m_LastResponse.AsArray()};
    for (auto& sub : arr)
    {
      if (sub.IsObject())
      {
        JSONPP::Object& object{sub.AsObject()};
        const auto iter{object.find(L"success")};
        if (iter != object.end())
        {
          if (iter->second.IsObject())
          {
            JSONPP::Object& sub2{iter->second.AsObject()};
            const auto iter2{sub2.find(L"username")};
            if (iter2 != sub2.end())
            {
              sClientKey = iter2->second.AsString();
              break;
            }
          }
        }
      }
    }

    return sClientKey.length() ? S_OK : ATL::AtlHresultFromWin32(ERROR_NOT_FOUND);
  }

  HRESULT DeleteUser(_In_z_ LPCSTR pszUsername, _In_z_ LPCSTR pszClientKey, _Inout_ std::vector<CErrorResponse>& errors)
  {
    std::wostringstream ss;
    ss << L"/api/";
    ss << pszUsername;
    ss << L"/config/whitelist/";
    ss << pszClientKey;
    std::string sResponse;
    HRESULT hr{Delete(ss.str().c_str(), sResponse)};
    if (FAILED(hr))
      return hr;
    try
    {
      m_LastResponse.Parse(sResponse.c_str());
    }
    catch (JSONPP::CParseException&)
    {
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
    }
    bool bErrorResponse{false};
    hr = IsErrorResponse(sResponse, bErrorResponse, errors);
    if (FAILED(hr) || bErrorResponse)
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
    return S_OK;
  }

  HRESULT GetConfiguration(_Inout_ std::vector<CErrorResponse>& errors)
  {
    //Validate our parameters
    ASSERT(m_sClientKey.length());

    std::wostringstream ss;
    ss << L"/api/";
    ss << m_sClientKey;
    ss << L"/config";
    std::wstring sURL{ss.str()};
    std::string sResponse;
    HRESULT hr{Get(ss.str().c_str(), sResponse)};
    if (FAILED(hr))
      return hr;
    try
    {
      m_LastResponse.Parse(sResponse.c_str());
    }
    catch (JSONPP::CParseException&)
    {
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
    }
    bool bErrorResponse{false};
    hr = IsErrorResponse(sResponse, bErrorResponse, errors);
    if (FAILED(hr) || bErrorResponse)
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
    return S_OK;
  }

  HRESULT GetFullState(_Inout_ std::vector<CErrorResponse>& errors)
  {
    //Validate our parameters
    ASSERT(m_sClientKey.length());

    std::wostringstream ss;
    ss << L"/api/";
    ss << m_sClientKey;
    std::wstring sURL{ss.str()};
    std::string sResponse;
    HRESULT hr{Get(ss.str().c_str(), sResponse)};
    if (FAILED(hr))
      return hr;
    try
    {
      m_LastResponse.Parse(sResponse.c_str());
    }
    catch (JSONPP::CParseException&)
    {
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
    }
    bool bErrorResponse{false};
    hr = IsErrorResponse(sResponse, bErrorResponse, errors);
    if (FAILED(hr) || bErrorResponse)
      return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
    return S_OK;
  }

  HRESULT Call(_In_z_ LPCWSTR pszURL, _In_z_ LPCWSTR pszVerb, _In_opt_z_ LPVOID pRequest, _In_ DWORD dwRequestSize, _Out_ std::string& sResponse)
  {
    //Issue the request
    WinHTTPWrappers::CSyncDownloader request;
    HRESULT hr{request.Initialize(m_Connection, pszURL, pszVerb)};
    if (FAILED(hr))
      return hr;
    if (dwRequestSize)
      hr = request.SendRequestSync(pRequest, dwRequestSize);
    else
      hr = request.SendRequestSync();
    if (FAILED(hr))
      return hr;

    //Update the response output parameter
    sResponse.clear();
#pragma warning(suppress: 26481 26486 26489 26490)
    sResponse.insert(sResponse.end(), reinterpret_cast<const char*>(request.m_Response.data()), reinterpret_cast<const char*>(request.m_Response.data()) + request.m_Response.size());
    return hr;
  }

  HRESULT Get(_In_z_ LPCWSTR pszURL, _Out_ std::string& sResponse)
  {
    return Call(pszURL, L"GET", nullptr, 0, sResponse);
  }

  HRESULT Post(_In_z_ LPCWSTR pszURL, _In_opt_z_ LPCSTR pszRequest, _Out_ std::string& sResponse)
  {
    if (pszRequest != nullptr)
    {
      std::string sCopyRequest{pszRequest};
#pragma warning(suppress: 26472)
      return Call(pszURL, L"POST", sCopyRequest.data(), static_cast<DWORD>(sCopyRequest.size()), sResponse);
    }
    else
      return Call(pszURL, L"POST", nullptr, 0, sResponse);
  }

  HRESULT Put(_In_z_ LPCWSTR pszURL, _In_z_ LPCSTR pszRequest, _Out_ std::string& sResponse)
  {
    if (pszRequest != nullptr)
    {
      std::string sCopyRequest{pszRequest};
#pragma warning(suppress: 26472)
      return Call(pszURL, L"PUT", sCopyRequest.data(), static_cast<DWORD>(sCopyRequest.size()), sResponse);
    }
    else
      return Call(pszURL, L"PUT", nullptr, 0, sResponse);
  }

  HRESULT Delete(_In_z_ LPCWSTR pszURL, _Out_ std::string& sResponse)
  {
    return Call(pszURL, L"DELETE", nullptr, 0, sResponse);
  }

//Member variables
  std::wstring m_sClientKey;
  WinHTTPWrappers::CConnection m_Connection;
  JSONPP::CValue m_LastResponse;
  JSONPP::CValue m_LastLights;
};

inline HRESULT CLight::Set(_In_ const CLightState& state, _Inout_ std::vector<CErrorResponse>& errors)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_pBridge != nullptr); //You need to use CBridge::GetLight to construct a CLight instance
#pragma warning(suppress: 26477)
  ATLASSERT(m_pBridge->m_sClientKey.length());
#pragma warning(suppress: 26477)
  ATLASSERT(m_nID != 0);

  std::ostringstream ss;
  ss << "{";
  if (state.m_On)
  {
    if (state.m_On.value())
      ss << "\"on\": true,";
    else
      ss << "\"on\": false,";
  }
  if (state.m_Brightness)
  {
    ss << "\"bri\": ";
    ss << static_cast<int>(state.m_Brightness.value());
    ss << ",";
  }
  if (state.m_Hue)
  {
    ss << "\"hue\": ";
    ss << state.m_Hue.value();
    ss << ",";
  }
  if (state.m_Saturation)
  {
    ss << "\"sat\": ";
    ss << static_cast<int>(state.m_Saturation.value());
    ss << ",";
  }
  if (state.m_XY)
  {
    ss << "\"xy\": [";
    ss << state.m_XY.value().first;
    ss << ", ";
    ss << state.m_XY.value().second;
    ss << "],";
  }
  if (state.m_CT)
  {
    ss << "\"ct\": ";
    ss << state.m_CT.value();
    ss << ",";
  }
  switch (state.m_Alert)
  {
    case CLightState::ALERT::NONE:
    {
      ss << R"("alert": "none",)";
      break;
    }
    case CLightState::ALERT::SELECT:
    {
      ss << R"("alert": "select",)";
      break;
    }
    case CLightState::ALERT::LSELECT:
    {
      ss << R"("alert": "lselect",)";
      break;
    }
    case CLightState::ALERT::NOT_SET: //deliberate fallthrough
    default:
    {
      break;
    }
  }
  switch (state.m_Effect)
  {
    case CLightState::EFFECT::NONE:
    {
      ss << R"("effect": "none",)";
      break;
    }
    case CLightState::EFFECT::COLORLOOP:
    {
      ss << R"("effect": "colorloop",)";
      break;
    }
    case CLightState::EFFECT::NOT_SET: //deliberate fallthrough
    default:
    {
      break;
    }
  }
  if (state.m_TransitionTime)
  {
    ss << "\"transitiontime\": ";
    ss << state.m_TransitionTime.value();
    ss << ",";
  }
  if (state.m_BrightnessInc)
  {
    ss << "\"bri_inc\": ";
    ss << state.m_BrightnessInc.value();
    ss << ",";
  }
  if (state.m_SaturationInc)
  {
    ss << "\"sat_inc\": ";
    ss << state.m_SaturationInc.value();
    ss << ",";
  }
  if (state.m_HueInc)
  {
    ss << "\"hue_inc\": ";
    ss << state.m_HueInc.value();
    ss << ",";
  }
  if (state.m_CTInc)
  {
    ss << "\"ct_inc\": ";
    ss << state.m_CTInc.value();
    ss << ",";
  }
  if (state.m_XYInc)
  {
    ss << "\"xy_inc\": [";
    ss << state.m_XYInc.value().first;
    ss << ", ";
    ss << state.m_XYInc.value().second;
    ss << "],";
  }
  std::string sRequest{ss.str()};
  sRequest = sRequest.substr(0, sRequest.length() - 1);
  sRequest += "}";

  std::wostringstream ss2;
  ss2 << L"/api/";
  ss2 << m_pBridge->m_sClientKey;
  ss2 << L"/lights/";
  ss2 << m_nID;
  ss2 << L"/state";
  std::string sResponse;
  HRESULT hr{m_pBridge->Put(ss2.str().c_str(), sRequest.c_str(), sResponse)};
  if (FAILED(hr))
    return hr;
  try
  {
    m_LastResponse.Parse(sResponse.c_str());
  }
  catch (JSONPP::CParseException&)
  {
    return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
  }
  bool bErrorResponse{false};
  hr = CBridge::IsErrorResponse(sResponse, bErrorResponse, errors);
  if (FAILED(hr) || bErrorResponse)
    return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
  if (!m_LastResponse.IsArray())
    return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
  return S_OK;
}

inline HRESULT CLight::Delete(_Inout_ std::vector<CErrorResponse>& errors)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_pBridge != nullptr); //You need to use CBridge::GetLight to construct a CLight instance
#pragma warning(suppress: 26477)
  ATLASSERT(m_pBridge->m_sClientKey.length());

  std::wostringstream ss;
  ss << L"/api/";
  ss << m_pBridge->m_sClientKey;
  ss << L"/lights/";
  ss << m_nID;
  std::string sResponse;
  HRESULT hr{m_pBridge->Delete(ss.str().c_str(), sResponse)};
  if (FAILED(hr))
    return hr;
  try
  {
    m_LastResponse.Parse(sResponse.c_str());
  }
  catch (JSONPP::CParseException&)
  {
    return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
  }
  bool bErrorResponse{false};
  hr = CBridge::IsErrorResponse(sResponse, bErrorResponse, errors);
  if (FAILED(hr) || bErrorResponse)
    return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
  return S_OK;
}

inline HRESULT CLight::Rename(_In_z_ LPCSTR pszNewName, _Inout_ std::vector<CErrorResponse>& errors)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_pBridge != nullptr); //You need to use CBridge::GetLight to construct a CLight instance
#pragma warning(suppress: 26477)
  ATLASSERT(m_pBridge->m_sClientKey.length());

  std::ostringstream ss;
  ss << R"({"name": ")";
  ss << pszNewName;
  ss << "\"}";
  std::wostringstream ss2;
  ss2 << L"/api/";
  ss2 << m_pBridge->m_sClientKey;
  ss2 << L"/lights/";
  ss2 << m_nID;
  std::string sResponse;
  HRESULT hr{m_pBridge->Put(ss2.str().c_str(), ss.str().c_str(), sResponse)};
  if (FAILED(hr))
    return hr;
  try
  {
    m_LastResponse.Parse(sResponse.c_str());
  }
  catch (JSONPP::CParseException&)
  {
    return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
  }
  bool bErrorResponse{false};
  hr = CBridge::IsErrorResponse(sResponse, bErrorResponse, errors);
  if (FAILED(hr) || bErrorResponse)
    return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
  return S_OK;
}

inline HRESULT CLight::Refresh(_Inout_ std::vector<CErrorResponse>& errors)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_pBridge != nullptr); //You need to use CBridge::GetLight to construct a CLight instance
#pragma warning(suppress: 26477)
  ATLASSERT(m_pBridge->m_sClientKey.length());

  std::wostringstream ss;
  ss << L"/api/";
  ss << m_pBridge->m_sClientKey;
  ss << L"/lights/";
  ss << m_nID;
  std::string sResponse;
  HRESULT hr{m_pBridge->Get(ss.str().c_str(), sResponse)};
  if (FAILED(hr))
    return hr;
  try
  {
    m_LastResponse.Parse(sResponse.c_str());
  }
  catch (JSONPP::CParseException&)
  {
    return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
  }
  bool bErrorResponse{false};
  hr = CBridge::IsErrorResponse(sResponse, bErrorResponse, errors);
  if (FAILED(hr) || bErrorResponse)
    return ATL::AtlHresultFromWin32(ERROR_WINHTTP_INVALID_SERVER_RESPONSE);
  m_LastState = m_LastResponse;
  return S_OK;
}


}; //namespace PHue++

#endif //#ifndef __PHUEPP_H__
